
import pytest
import cvxpy as cp
import numpy as np
import glob
import os
import importlib
import itertools
import pickle
import utils_test
import sys
sys.path.append('../')
from cvxpygen import cpg


def actuator_problem():

    # define dimensions (test for degenerate vectors and matrices)
    n, m = 1, 1

    # define variables
    u = cp.Variable(n, name='u')
    delta_u = cp.Variable((n, n), name='delta_u')

    # define parameters
    A = cp.Parameter((m, n), name='A')
    w = cp.Parameter(m, name='w')
    lamb_sm = cp.Parameter(nonneg=True, name='lamb_sm')
    kappa = cp.Parameter(n, nonneg=True, name='kappa')
    u_prev = cp.Parameter(n, name='u_prev')
    u_min = cp.Parameter(n, name='u_min')
    u_max = cp.Parameter(n, name='u_max')

    # define objective
    objective = cp.Minimize(cp.sum_squares(A @ u - w) + lamb_sm * cp.sum_squares(delta_u) + kappa @ cp.abs(u))

    # define constraints
    constraints = [u_min <= u, u <= u_max, delta_u == u - u_prev]

    # define problem
    return cp.Problem(objective, constraints)


def MPC_problem():

    # define dimensions
    H, n, m = 10, 6, 3

    # define variables
    U = cp.Variable((m, H), name='U')
    X = cp.Variable((n, H + 1), name='X')

    # define parameters
    Psqrt = cp.Parameter((n, n), name='Psqrt', diag=True)
    Qsqrt = cp.Parameter((n, n), name='Qsqrt', diag=True)
    Rsqrt = cp.Parameter((m, m), name='Rsqrt', diag=True)
    A = cp.Parameter((n, n), name='A', sparsity=[(i, i) for i in range(n)] + [(i, 3+i) for i in range(n // 2)])
    B = cp.Parameter((n, m), name='B', sparsity=[(3+i, i) for i in range(n // 2)])
    x_init = cp.Parameter(n, name='x_init')

    # define objective
    objective = cp.Minimize(
        cp.sum_squares(Psqrt @ X[:, H - 1]) + cp.sum_squares(Qsqrt @ X[:, :H]) + cp.sum_squares(Rsqrt @ U))

    # define constraints
    constraints = [X[:, 1:] == A @ X[:, :H] + B @ U,
                   cp.abs(U) <= 1,
                   X[:, 0] == x_init]

    # define problem
    return cp.Problem(objective, constraints)


def portfolio_problem():

    # define dimensions
    n, m = 100, 10

    # define variables
    w = cp.Variable(n, name='w')
    delta_w = cp.Variable(n, name='delta_w')
    f = cp.Variable(m, name='f')

    # define parameters
    a = cp.Parameter(n, name='a')
    F = cp.Parameter((n, m), name='F')
    Sig_f_sqrt = cp.Parameter((m, m), name='Sig_f_sqrt')
    d_sqrt = cp.Parameter(n, name='d_sqrt')
    k_tc = cp.Parameter(n, nonneg=True, name='k_tc')
    k_sh = cp.Parameter(n, nonneg=True, name='k_sh')
    w_prev = cp.Parameter(n, name='w_prev')
    L = cp.Parameter(nonneg=True, name='L')

    # define objective
    objective = cp.Maximize(a @ w
                            - cp.sum_squares(Sig_f_sqrt @ f)
                            - cp.sum_squares(cp.multiply(d_sqrt, w))
                            - k_tc @ cp.abs(delta_w)
                            + k_sh @ cp.minimum(0, w))

    # define constraints
    constraints = [f == F.T @ w,
                   np.ones(n) @ w == 1,
                   cp.norm(w, 1) <= L,
                   delta_w == w - w_prev]

    # define problem
    return cp.Problem(objective, constraints)


def assign_data(prob, name, seed):

    np.random.seed(seed)

    if name == 'actuator':

        prob.param_dict['A'].value = np.array([[1]])
        prob.param_dict['w'].value = np.array([1])
        prob.param_dict['lamb_sm'].value = np.random.rand()
        prob.param_dict['kappa'].value = 0.1 * np.ones(1)
        prob.param_dict['u_prev'].value = 0 * np.ones(1)
        prob.param_dict['u_min'].value = -1 * np.ones(1)
        prob.param_dict['u_max'].value = 1 * np.ones(1)

    elif name == 'MPC':

        # continuous-time dynmaics
        A_cont = np.concatenate((np.array([[0, 0, 0, 1, 0, 0],
                                           [0, 0, 0, 0, 1, 0],
                                           [0, 0, 0, 0, 0, 1]]),
                                           np.zeros((3, 6))), axis=0)
        mass = 1
        B_cont = np.concatenate((np.zeros((3, 3)),
                                 (1 / mass) * np.diag(np.ones(3))), axis=0)

        # discrete-time dynamics
        td = 0.1

        prob.param_dict['A'].value = np.eye(6) + td * A_cont
        prob.param_dict['B'].value = td * B_cont
        prob.param_dict['Psqrt'].value = np.eye(6)
        prob.param_dict['Qsqrt'].value = np.eye(6)
        prob.param_dict['Rsqrt'].value = np.sqrt(0.1) * np.eye(3)
        prob.param_dict['x_init'].value = -2*np.ones(6) + 4*np.random.rand(6)

    elif name == 'portfolio':

        n, m = 100, 10
        gamma = 1
        alpha = np.random.randn(n)
        kappa_tc = 0.01 * np.ones(n)
        kappa_sh = 0.05 * np.ones(n)
        prob.param_dict['a'].value = alpha / gamma
        prob.param_dict['F'].value = np.round(np.random.randn(n, m))
        prob.param_dict['Sig_f_sqrt'].value = np.diag(np.random.rand(m))
        prob.param_dict['d_sqrt'].value = np.random.rand(n)
        prob.param_dict['k_tc'].value = kappa_tc / gamma
        prob.param_dict['k_sh'].value = kappa_sh / gamma
        prob.param_dict['w_prev'].value = np.zeros(n)
        prob.param_dict['L'].value = 1.6

    return prob


def get_primal_vec(prob, name):
    if name == 'actuator':
        return np.array((prob.var_dict['u'].value.squeeze(), prob.var_dict['delta_u'].value.squeeze()))
    elif name == 'MPC':
        return np.concatenate((prob.var_dict['U'].value.flatten(), prob.var_dict['X'].value.flatten()))
    elif name == 'portfolio':
        return np.concatenate((prob.var_dict['w'].value, prob.var_dict['delta_w'].value, prob.var_dict['f'].value))


N_RAND = 3

name_solver_style_seed = [['actuator', 'MPC', 'portfolio'],
                          ['OSQP', 'SCS'],
                          ['unroll', 'loops'],
                          list(np.arange(N_RAND))]

name_to_prob = {'actuator': actuator_problem(),
                'MPC': MPC_problem(),
                'portfolio': portfolio_problem()}


@pytest.mark.parametrize('name, solver, style, seed', list(itertools.product(*name_solver_style_seed)))
def test(name, solver, style, seed):

    prob = name_to_prob[name]

    if seed == 0:
        prob = assign_data(prob, name, 0)
        if style == 'unroll':
            cpg.generate_code(prob, code_dir='test_%s_%s_unroll' % (name, solver), solver=solver, unroll=True,
                              prefix='%s_%s_ex' % (name, solver))
            assert len(glob.glob(os.path.join('test_%s_%s_unroll' % (name, solver), 'cpg_module.*'))) > 0
        if style == 'loops':
            cpg.generate_code(prob, code_dir='test_%s_%s_loops' % (name, solver), solver=solver, unroll=False,
                              prefix='%s_%s_im' % (name, solver))
            assert len(glob.glob(os.path.join('test_%s_%s_loops' % (name, solver), 'cpg_module.*'))) > 0

    with open('test_%s_%s_%s/problem.pickle' % (name, solver, style), 'rb') as f:
        prob = pickle.load(f)

    module = importlib.import_module('test_%s_%s_%s.cpg_solver' % (name, solver, style))
    prob.register_solve('CPG', module.cpg_solve)

    prob = assign_data(prob, name, seed)

    val_py, prim_py, dual_py, val_cg, prim_cg, dual_cg, prim_py_norm, dual_py_norm = \
        utils_test.check(prob, solver, name, get_primal_vec)

    if not np.isinf(val_py):
        assert abs((val_cg - val_py) / val_py) < 0.1

    if prim_py_norm > 1e-6:
        assert np.linalg.norm(prim_cg - prim_py, 2) / prim_py_norm < 0.1
    else:
        assert np.linalg.norm(prim_cg, 2) < 1e-3

    if dual_py_norm > 1e-6:
        assert np.linalg.norm(dual_cg - dual_py, 2) / dual_py_norm < 0.1
    else:
        assert np.linalg.norm(dual_cg, 2) < 1e-3




import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np

n, m = 4, 5

nodes_pos = np.array([[0, 0],
                      [1, 0],
                      [1, 1],
                      [2, 0]])

edges_ind = np.array([[0, 1],
                      [0, 2],
                      [1, 2],
                      [1, 3],
                      [2, 3]])

fig, (ax, ax_w) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [5, 1]})
fig.set_size_inches(10, 5)

col_pipes = '#222222'
col_flows = ['b', 'm', 'c', 'y']
width_fac = 0.06

ax.axis('equal')
ax.axis('off')


def create_animation(prob, name):

    prob.param_dict['R'].value = np.array([[1, 0, 1, 0],
                                           [0, 1, 0, 1],
                                           [0, 0, 1, 1],
                                           [1, 0, 0, 1],
                                           [0, 1, 1, 0]])
    prob.param_dict['f_min'].value = np.zeros(n)
    prob.param_dict['f_max'].value = np.ones(n)
    prob.param_dict['c'].value = np.array([0.5, 1, 1, 1, 0.7])

    def init():

        ret = []

        for i in range(nodes_pos.shape[0]):
            circle = plt.Circle(tuple(nodes_pos[i, :]), 0.15, color=col_pipes, clip_on=False)
            ax.add_patch(circle)

        for i in range(edges_ind.shape[0]):

            ret.append(ax.plot([], []))
            ret.append(ax.plot([], []))

            for j in np.argwhere(prob.param_dict['R'].value[i, :]).flatten():
                poly = plt.Polygon(np.zeros((1, 2)))
                ax.add_patch(poly)
                ret.append(poly)

        ret.append(ax_w.bar(0, 1))

        return ax, ax_w, ret

    def update(frame):

        ret = []

        prob.param_dict['w'].value = np.array([1 + 3 * (0.5 + 0.5 * np.sin(frame * 1.9)),
                                               1 + 3 * (0.5 + 0.5 * np.sin(frame * 1.4)),
                                               1 + 3 * (0.5 + 0.5 * np.sin(frame)),
                                               1 + 3 * (0.5 + 0.5 * np.sin(frame * 0.8))])
        prob.solve(method='CPG')

        ax.clear()
        ax.axis('equal')
        ax.axis('off')
        ax_w.clear()
        ax_w.axis('off')

        for i in range(nodes_pos.shape[0]):
            circle = plt.Circle(tuple(nodes_pos[i, :]), 0.15, color=col_pipes, clip_on=False)
            ax.add_patch(circle)
            ret.append(circle)

        for i in range(edges_ind.shape[0]):

            fill = 0
            start = nodes_pos[edges_ind[i, 0], :]
            end = nodes_pos[edges_ind[i, 1], :]
            vector = end - start
            offset_raw = np.array([vector[1], -vector[0]])
            offset_norm = offset_raw / np.linalg.norm(offset_raw)
            offset = width_fac * prob.param_dict['c'].value[i] * offset_norm
            x_values = np.array([start[0], end[0]])
            y_values = np.array([start[1], end[1]])

            ret.append(ax.plot(x_values + offset[0], y_values + offset[1], color=col_pipes))
            ret.append(ax.plot(x_values - offset[0], y_values - offset[1], color=col_pipes))

            for j in np.argwhere(prob.param_dict['R'].value[i, :]).flatten():

                fill_ratio = prob.var_dict['f'].value[j] / prob.param_dict['c'].value[i]

                offset1 = (1 - 2 * fill) * offset
                offset2 = offset1 - 2 * fill_ratio * offset

                poly = plt.Polygon(np.array([[x_values[0] + offset1[0], y_values[0] + offset1[1]],
                                             [x_values[1] + offset1[0], y_values[1] + offset1[1]],
                                             [x_values[1] + offset2[0], y_values[1] + offset2[1]],
                                             [x_values[0] + offset2[0], y_values[0] + offset2[1]]]),
                                   color=col_flows[j], zorder=-1)
                ax.add_patch(poly)
                ret.append(poly)

                fill += fill_ratio

        bar = ax_w.bar(np.arange(prob.param_dict['w'].size), prob.param_dict['w'].value)

        for j, b in enumerate(bar):
            b.set_color(col_flows[j])

        ret.append(bar)
        ax_w.set_ylim(0, 4)
        ax_w.set_title('w')
        ax_w.axis('off')

        return ax, ax_w, ret

    ani = FuncAnimation(fig, update, frames=np.linspace(0, 2 * np.pi, 128), init_func=init, blit=False)
    ani.save('%s.gif' % name)



import numpy as np
import matplotlib.pyplot as plt
from numpy import radians as rad
from matplotlib.patches import Arc, RegularPolygon
from matplotlib.animation import FuncAnimation


def get_circ(thet):
    if thet > 0:
        thet1 = 0
        thet2 = thet
        end = thet
        ori = thet
    else:
        thet1 = 360 + thet
        thet2 = 360
        end = 360 + thet
        ori = 180 + thet

    arc = Arc((0.0, 0.0), 1, 1, angle=0, theta1=thet1, theta2=thet2, capstyle='round', linestyle='-', lw=2, color='blue')

    endX = 0.5 * np.cos(rad(end))
    endY = 0.5 * np.sin(rad(end))

    pol = RegularPolygon((endX, endY), 3, 0.1, rad(ori), color='blue')

    return arc, pol


def init():
    arr_bl = plt.arrow([], [], [], [])
    arr_br = plt.arrow([], [], [], [])
    arr_tr = plt.arrow([], [], [], [])
    arr_tl = plt.arrow([], [], [], [])
    circ_t = Arc((0, 0), 0, 0)
    head_t = RegularPolygon((0, 0), 3)
    return arr_bl, arr_br, arr_tr, arr_tl, circ_t, head_t


def create_animation(prob, name):
    
    fig, ax = plt.subplots()
    arr_width = .02
    kw = dict(width=arr_width, head_width=3 * arr_width, head_length=9 * arr_width, color='r')

    def update(frame):
        w_value = np.array([0.8 * np.cos(frame), 0.8 * np.sin(frame), np.sin(1.4 * frame)])
        prob.param_dict['w'].value = w_value
        prob.param_dict['u_prev'].value = prob.var_dict['u'].value
        prob.solve(method='CPG')
        u_value = prob.var_dict['u'].value

        ax.clear()
        ax.axis('equal')
        ax.set_xlim(-3, 3)
        ax.set_ylim(-3, 3)
        ax.axis('off')
        ax.add_patch(plt.Rectangle((-1, -1), 2, 2, facecolor='w', edgecolor='k'))

        arr_bl = plt.arrow(-1, -1, u_value[0], u_value[1], **kw)
        arr_br = plt.arrow(1, -1, u_value[2], u_value[3], **kw)
        arr_tr = plt.arrow(1, 1, u_value[4], u_value[5], **kw)
        arr_tl = plt.arrow(-1, 1, u_value[6], u_value[7], **kw)
        arr_w = plt.arrow(0, 0, w_value[0], w_value[1], **kw)

        circ_t, head_t = get_circ(340 * w_value[2])
        ax.add_patch(circ_t)
        ax.add_patch(head_t)

        return arr_bl, arr_br, arr_tr, arr_tl, arr_w, circ_t, head_t

    ani = FuncAnimation(fig, update, frames=np.linspace(0, 2 * np.pi, 128), init_func=init, blit=True)

    ani.save('%s.gif' % name)




import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import numpy as np


fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, gridspec_kw={'height_ratios': [3, 0.1]})
fig.set_size_inches(4, 5)


def init():

    map_X = ax1.imshow(np.zeros((1, 1)))
    map_W = ax2.imshow(np.zeros((1, 1)))
    map_r = ax3.imshow(np.zeros((1, 1)))

    return map_X, map_W, map_r


def create_animation(prob, name):

    kwargs = dict(cmap='jet', vmin=0, vmax=1)

    def update(frame):

        for i in range(prob.param_dict['r'].shape[0]):
            prob.param_dict['r'].value[i] = np.sin((1 - 0.05 * i) * frame)

        prob.solve(method='CPG')

        map_X = ax1.imshow(prob.var_dict['X'].value, **kwargs)
        map_W = ax2.imshow(prob.param_dict['W'].value, **kwargs)
        map_r = ax3.imshow(prob.param_dict['r'].value.reshape(1, 10), **kwargs)

        ax1.set_title('X')
        ax2.set_title('W')
        ax3.set_title('r')

        if frame == 0:
            plt.colorbar(map_W, ax=ax2)

        for ax in [ax1, ax2, ax3, ax4]:
            ax.axis('off')

        plt.tight_layout()

        return map_X, map_W, map_r

    ani = FuncAnimation(fig, update, frames=np.linspace(0, 2 * np.pi, 24), init_func=init, blit=True)
    ani.save('%s.gif' % name)




import cvxpy as cp
from cvxpygen import cpg
import numpy as np
import time
import sys

if __name__ == "__main__":

    '''
    1. Generate Code
    '''

    # define CVXPY problem
    m, n = 3, 2
    x = cp.Variable(n, name='x')
    A = cp.Parameter((m, n), name='A', sparsity=[(0, 0), (0, 1), (1, 1)])
    b = cp.Parameter(m, name='b')
    problem = cp.Problem(cp.Minimize(cp.sum_squares(A @ x - b)), [x >= 0])

    # assign parameter values and test-solve
    np.random.seed(0)
    A.value = np.zeros((m, n))
    A.value[0, 0] = np.random.randn()
    A.value[0, 1] = np.random.randn()
    A.value[1, 1] = np.random.randn()
    b.value = np.random.randn(m)
    problem.solve(solver='SCS')

    # generate code
    cpg.generate_code(problem, code_dir='nonneg_LS', solver='SCS')

    '''
    2. Solve & Compare
    '''

    # import extension module and register custom CVXPY solve method
    from nonneg_LS.cpg_solver import cpg_solve
    problem.register_solve('cpg', cpg_solve)

    # solve problem conventionally
    t0 = time.time()
    val = problem.solve(solver='SCS')
    t1 = time.time()
    sys.stdout.write('\nCVXPY\nSolve time: %.3f ms\n' % (1000*(t1-t0)))
    sys.stdout.write('Primal solution: x = [%.6f, %.6f]\n' % tuple(x.value))
    sys.stdout.write('Dual solution: d0 = [%.6f, %.6f]\n' % tuple(problem.constraints[0].dual_value))
    sys.stdout.write('Objective function value: %.6f\n' % val)

    # solve problem with C code via python wrapper
    t0 = time.time()
    val = problem.solve(method='cpg', updated_params=['A', 'b'], verbose=False)
    t1 = time.time()
    sys.stdout.write('\nCVXPYgen\nSolve time: %.3f ms\n' % (1000 * (t1 - t0)))
    sys.stdout.write('Primal solution: x = [%.6f, %.6f]\n' % tuple(x.value))
    sys.stdout.write('Dual solution: d0 = [%.6f, %.6f]\n' % tuple(problem.constraints[0].dual_value))
    sys.stdout.write('Objective function value: %.6f\n' % val)




import numpy as np


def get_dual_vec(prob):
    dual_values = []
    for constr in prob.constraints:
        if constr.args[0].size == 1:
            dual_values.append(np.atleast_1d(constr.dual_value).flatten())
        else:
            dual_values.append(constr.dual_value.flatten())
    return np.concatenate(dual_values)


def check(prob, solver, name, func_get_primal_vec):

    if solver == 'OSQP':
        val_py = prob.solve(solver='OSQP', eps_abs=1e-3, eps_rel=1e-3, max_iter=4000, polish=False,
                            adaptive_rho_interval=int(1e6), warm_start=False)
    elif solver == 'SCS':
        val_py = prob.solve(solver='SCS', warm_start=False, verbose=False)
    else:
        val_py = prob.solve(solver='ECOS')
    prim_py = func_get_primal_vec(prob, name)
    dual_py = get_dual_vec(prob)
    if solver == 'OSQP':
        val_cg = prob.solve(method='CPG', warm_start=False)
    elif solver == 'SCS':
        val_cg = prob.solve(method='CPG', warm_start=False, verbose=False)
    else:
        val_cg = prob.solve(method='CPG')
    prim_cg = func_get_primal_vec(prob, name)
    dual_cg = get_dual_vec(prob)
    prim_py_norm = np.linalg.norm(prim_py, 2)
    dual_py_norm = np.linalg.norm(dual_py, 2)

    return val_py, prim_py, dual_py, val_cg, prim_cg, dual_cg, prim_py_norm, dual_py_norm
